# Copyright (c) 2013..2015 Dirk Jagdmann <doj@cubic.org>

# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.

# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:

#     1. The origin of this software must not be misrepresented; you
#        must not claim that you wrote the original software. If you use
#        this software in a product, an acknowledgment in the product
#        documentation would be appreciated but is not required.

#     2. Altered source versions must be plainly marked as such, and
#        must not be misrepresented as being the original software.

#     3. This notice may not be removed or altered from any source
#        distribution.

# $Id: DB.pm 7734 2016-04-10 17:33:29Z doj $

package PayPal::DB;

use Fcntl; # For O_RDWR, O_CREAT, etc.
use SDBM_File;
use JSON;
use File::Path;
use File::Find;
use File::HomeDir;
use Carp;
use strict;
use lib '../';
use configuration;

##############################################################################
# misc
##############################################################################

sub number2cent
{
    local $_ = shift;
    s/,//g;	# strip comma characters
    $_ .= '.00' unless /\.\d+$/;
    $_ .= '0' if /\.\d$/; # append digit if only 10c
    s/\.(\d\d)\d?/$1/;
    s/\.//g;	# strip point characters
    return $_;
}

BEGIN
{
    number2cent('1') eq '100' or die;
    number2cent('13') eq '1300' or die;
    number2cent('1.2') eq '120' or die;
    number2cent('1.23') eq '123' or die;
    number2cent('1.234') eq '123' or die;
    number2cent('12345.6') eq '1234560' or die;
}

##############################################################################
# txn db
##############################################################################

# generate the PayPal DB directory name.
# @return directory name of PayPal DB.
my $paypal_db_dir_name = sub
{
    my $stem = $configuration::paypal_login;
    die '$configuration::paypal_login not defined' unless length($stem);
    $stem =~ s/\W/_/g;		# replace any non-word characters
    die "\$configuration::paypal_login does not contain word characters: $configuration::paypal_login" unless length($stem);
    return File::HomeDir->my_data . "/.paypal_$stem/";
};

# generate the PayPal DB directory name and create the directory.
# @return directory name of PayPal DB.
our $paypal_db_dir = sub
{
    my $dir = $paypal_db_dir_name->();
    unless(-d $dir) {
	File::Path::make_path($dir) or croak "could not mkdir $dir";
    }
    return $dir;
};

# generate file name for a PayPal transaction.
# @param $txn transaction ID.
# @return file name for transaction $txn.
my $paypal_txn_fn = sub
{
    my ($txn) = @_;
    return unless length($txn) > 0;
    return unless $txn =~ /^[a-zA-Z0-9-]+$/;
    my $fn = $txn;
    my $dir = $paypal_db_dir->();
    if (length($txn) == 1) {
	$fn = "_$fn";
    } else {
	$dir .= substr($txn, 0, 1) . '/' . substr($txn, 1, 1);
    }
    unless(-d $dir) {
	File::Path::make_path($dir) or croak "could not mkdir $dir";
    }
    return "${dir}/${fn}";
};

CHECK {
    if ($paypal_txn_fn) {
	die unless $paypal_txn_fn->('O-16J9521443099411F');
    }
}

sub save_txn
{
    my ($h) = @_;
    # check parameter
    unless (ref($h) eq 'HASH') { carp "hash reference expected as argument"; return }
    # check if we have a useful transaction ID
    unless (exists($h->{'TransactionID'})) { carp "TransactionID does not exist"; return }
    my $txn = $h->{'TransactionID'};
    unless (length($txn) > 0) { carp "TransactionID is empty string"; return }

    # check signs on Gross, Net, Fee
    foreach my $k (qw(GrossAmount NetAmount FeeAmount)) {
	$h->{$k} =~ s/,//g;	# strip comma characters
	unless ($h->{$k} =~ /^[+-]?\d+(\.\d{1,2})?$/) { carp "$k is not a number"; return }
    }
    if ($h->{GrossAmount} > 0.0 && $h->{NetAmount} > 0.0 && $h->{FeeAmount} <= 0.0) {
	# ok
    } elsif ($h->{GrossAmount} < 0.0 && $h->{NetAmount} < 0.0 && $h->{FeeAmount} >= 0.0) {
	# ok
    } else {
	carp "Gross $h->{GrossAmount} and Net $h->{NetAmount} and Fee $h->{FeeAmount} are strange";
	return;
    }
    my $g = number2cent($h->{GrossAmount});
    my $f = number2cent($h->{FeeAmount});
    my $n = number2cent($h->{NetAmount});
    unless ($g + $f == $n) {
	carp "Gross $h->{GrossAmount} + Fee $h->{FeeAmount} != Net $h->{NetAmount}, $g + $f == $n";
	return
    }

    # only copy useful values into %m
    my %m;
    while(my ($k,$v) = each(%$h)) {
	$m{$k} = $v if length($v) > 0;
    }
    # save %m
    my $fn = $paypal_txn_fn->($txn);
    unless ($fn) {
	carp "could not create filename for transaction $txn";
	return;
    }
    my $json = to_json(\%m, { utf8 => 1, pretty => 1 } );
    #print STDERR "PayPal::DB::save($txn) $json\n";
    open(A, ">$fn") or croak "could not open $fn";
    print A $json;
    close(A);
    return $txn;
}

sub get_txn
{
    my ($txn) = @_;
    return unless length($txn) > 0;
    my $fn = $paypal_txn_fn->($txn);
    unless ($fn) {
	carp "could not create filename for transaction $txn";
	return
    }
    local $/ = undef; # slurp whole file
    unless (open(A, $fn)) {
	return
    }
    my $json = <A>;
    close(A);
    return unless length($json) > 0;
    my $h = from_json($json, { utf8  => 1 } );
    if ($txn ne $h->{TransactionID}) {
	carp "txn db error: $txn != $h->{TransactionID}";
	return
    }
    return $h;
}

sub dump_txn
{
    local $/ = undef; # slurp whole file
    if (-f $_[0])
    {
	open(A, $_[0]) or croak "could not open $_[0]";
	my $rec = <A>;
	my $txn = from_json($rec, { utf8 => 1 } );
	print to_json($txn, { utf8 => 1, pretty => 1 } );
	close(A);
	return 1;
    }
    if (length($_[0]) > 0)
    {
	my $fn = $paypal_txn_fn->($_[0]);
	if (-f $fn) {
	    return dump_txn($fn);
	}
	carp "could not create filename for $_[0]";
	return
    }
    my $z = 0;
    print "{\n";
    find(sub {
        return unless -f $_;
	return if /\.(dir|pag)$/; # skip tied hash files
	print "\"$_\" : ";
	dump_txn($_);
	print ",\n";
	++$z;
    }, $paypal_db_dir->());
    print "\"total number of transactions\" : $z}\n";
    return $z;
}

##############################################################################
# state db
##############################################################################

my $state_fn = $paypal_db_dir->() . "state";

# open state file
my %state;
tie(%state, 'SDBM_File', $state_fn, O_RDWR|O_CREAT, 0666) or die "Couldn't tie SDBM file $state_fn: $!";
END {untie %state}

sub del_all_states
{
    print "deleting transactions from state file $state_fn\n";
    %state = ();
    return 1;
}

sub get_state
{
    my ($k) = @_;
    return unless length($k) > 0;
    return unless exists $state{$k};
    return $state{$k};
}

sub set_state
{
    my ($k, $v) = @_;
    return unless length($k) > 0;
    $state{$k} = $v;
    return 1;
}

sub set_states
{
    my $v = shift;
    foreach my $k (@_) {
	$state{$k} = $v;
    }
    return 1;
}

##############################################################################
# command line
##############################################################################

# strings used to parse command line
my $arg__txn = '--txn';
my $arg__delstate = '--delstate';
my $arg__paypaldb_dump = '--paypaldb_dump';
my $fetch_me = 'fetch me';

sub parse_command_line
{
    my $txn;
    foreach my $arg (@_) {
	if ($txn eq $fetch_me) {
	    exit ! dump_txn($arg);
	}
	if ($arg eq $arg__txn) {
	    $txn = $fetch_me;
	    next;
	}

	if ($arg eq $arg__delstate) {
	    del_all_states();
	    exit 0;
	}

	if ($arg eq $arg__paypaldb_dump) {
	    dump_txn();
	    exit 0;
	}
    }
    die command_line_help() if $txn eq $fetch_me;
}

sub command_line_help
{
    my ($prog) = @_;
    my $db_dir_name = $paypal_db_dir_name->();
  print<<"EOT";

usage: $prog $arg__delstate
  Delete the transaction state from previous file conversions.
  The $prog program may convert transactions a second time,
  which may lead to transactions occuring multiple times in the output file.

usage: $prog $arg__paypaldb_dump
  Dump all transactions from PayPal database.

usage: $prog $arg__txn <id>
  Show transaction <id>.

Transactions are saved in '$db_dir_name'
EOT
}

1;
